@@ -231,8 +231,9 @@ enum ElementState {
 	Name,
 	Eq,
 	Close,
-	/// Delimiter and Alphabet
-	AttributeValue(char, CodepointRanges),
+	/// Delimiter, Alphabet and whether we just read a CR, because of the mess
+	/// which is CRLF -> LF normalization.
+	AttributeValue(char, CodepointRanges, bool),
 	/// Encountered ?
 	MaybeXMLDeclEnd,
 	/// Encountered /
@@ -270,6 +271,11 @@ enum ContentState {
 	/// `]]>` sequence, either within cdata (true) or not (false)
 	/// if not within cdata, encountering this sequence is illegal
 	MaybeCDataEnd(bool, usize),
+	/// `\r` read, we need to look ahead by one char to see if it is a `\n`
+	/// before substituting
+	///
+	/// bool indicates whether we’re in a cdata section, because yes, this also applies to those
+	MaybeCRLF(bool),
 }
 
 
@@ -284,7 +290,7 @@ impl RefReturnState {
 		match self {
 			Self::AttributeValue(kind, delim, selector) => State::Element{
 				kind: kind,
-				state: ElementState::AttributeValue(delim, selector),
+				state: ElementState::AttributeValue(delim, selector, false),
 			},
 			Self::Text => State::Content(ContentState::Initial),
 		}
@@ -372,7 +378,7 @@ impl LexerOptions {
 	/// The defaults are implementation-defined and should not be relied upon.
 	pub fn defaults() -> LexerOptions {
 		LexerOptions{
-			max_token_length: 65535,
+			max_token_length: 8192,
 		}
 	}
 
@@ -461,7 +467,7 @@ pub struct Lexer {
 	prev_state: (String, State),
 	#[cfg(debug_assertions)]
 	last_single_read: Option<Utf8Char>,
-	err: Option<Error>,
+	err: Option<Box<Error>>,
 }
 impl Lexer {
 	/// Construct a new Lexer based on [`LexerOptions::defaults()`].
@@ -494,12 +500,21 @@ impl Lexer {
 		self.last_token_end = self.ctr.wrapping_sub(without);
 	}
 
+	#[inline]
+	fn prep_scratchpad(&mut self) {
+		if self.scratchpad.capacity() < self.opts.max_token_length {
+			// unless there is a bug, we should never exceed the capacity requested by max_token_length, so we go with reserve_exact
+			self.scratchpad.reserve_exact(self.opts.max_token_length - self.scratchpad.capacity())
+		}
+	}
+
 	fn read_validated<'r, 'x, R: CodepointRead, S: CharSelector>(&mut self, r: &'r mut R, selector: &'x S, limit: usize) -> Result<Endpoint> {
 		let remaining = match limit.checked_sub(self.scratchpad.len()) {
 			None => return Ok(Endpoint::Limit),
 			Some(v) => v,
 		};
 		let old_len = self.scratchpad.len();
+		self.prep_scratchpad();
 		let result = read::read_validated(
 			r,
 			selector,
@@ -532,7 +547,7 @@ impl Lexer {
 		&mut self,
 		r: &'r mut R,
 		selector: &'s S,
-		) -> Result<(usize, Endpoint)>
+		) -> (usize, Result<Endpoint>)
 	{
 		let (nread, result) = read::skip_matching(r, selector);
 		self.ctr = self.ctr.wrapping_add(nread);
@@ -541,9 +556,9 @@ impl Lexer {
 				if let Endpoint::Delimiter(ch) = ep {
 					self.ctr = self.ctr.wrapping_add(ch.len())
 				};
-				Ok((nread, ep))
+				(nread, Ok(ep))
 			},
-			Err(e) => Err(e),
+			Err(e) => (nread, Err(e)),
 		}
 	}
 
@@ -573,34 +588,36 @@ impl Lexer {
 		}
 	}
 
-	fn flush_scratchpad(&mut self) -> String {
-		let result = self.scratchpad.split_off(0);
-		debug_assert!(self.scratchpad.len() == 0);
+	fn flush_scratchpad<U, T: FnOnce(&str) -> U>(&mut self, conv: T) -> U {
+		let result = conv(&self.scratchpad);
+		self.scratchpad.clear();
 		result
 	}
 
 	fn flush_scratchpad_as_name(&mut self) -> Name {
-		let result = self.flush_scratchpad();
-		#[cfg(debug_assertions)]
-		{
-			return Name::from_string(result).unwrap();
-		}
-		#[cfg(not(debug_assertions))]
-		unsafe {
-			return Name::from_string_unchecked(result);
-		}
+		self.flush_scratchpad(|s| {
+			#[cfg(debug_assertions)]
+			{
+				Name::from_str(s).unwrap()
+			}
+			#[cfg(not(debug_assertions))]
+			unsafe {
+				Name::from_str_unchecked(s)
+			}
+		})
 	}
 
 	fn flush_scratchpad_as_cdata(&mut self) -> CData {
-		let result = self.flush_scratchpad();
-		#[cfg(debug_assertions)]
-		{
-			return CData::from_string(result).unwrap();
-		}
-		#[cfg(not(debug_assertions))]
-		unsafe {
-			return CData::from_string_unchecked(result);
-		}
+		self.flush_scratchpad(|s| {
+			#[cfg(debug_assertions)]
+			{
+				CData::from_str(s).unwrap()
+			}
+			#[cfg(not(debug_assertions))]
+			unsafe {
+				CData::from_str_unchecked(s)
+			}
+		})
 	}
 
 	fn maybe_flush_scratchpad_as_text(&mut self, without: usize) -> Option<Token> {
@@ -652,6 +669,13 @@ impl Lexer {
 					tok,
 				)))
 			},
+			'\r' => {
+				// CRLF needs to be folded to LF, and standalone LF needs, too
+				Ok(Some(ST(
+					State::Content(ContentState::MaybeCRLF(false)),
+					None,
+				)))
+			},
 			_ => Ok(None),
 		}
 	}
@@ -686,6 +710,7 @@ impl Lexer {
 					},
 					ch => if CLASS_XML_NAMESTART.select(ch) {
 						// add the first character to the scratchpad, because read_single does not do that
+						self.prep_scratchpad();
 						self.scratchpad.push_str(utf8ch.as_str());
 						Ok(ST(
 							State::Element{
@@ -750,6 +775,27 @@ impl Lexer {
 		}
 	}
 
+	fn lex_resume_text(&mut self, utf8ch: Utf8Char) -> Result<ST> {
+		let ch = utf8ch.to_char();
+		match self.lex_posttext_char(utf8ch)? {
+			// special delimiter char -> state transition
+			Some(st) => Ok(st),
+			// no special char -> check if it is valid text and proceed accordingly
+			None => if CLASS_XML_NONCHAR.select(ch) {
+				// non-Char, error
+				Err(Error::NotWellFormed(WFError::InvalidChar(ERRCTX_TEXT, ch as u32, false)))
+			} else {
+				// nothing special, push to scratchpad and return to initial content state
+				self.prep_scratchpad();
+				self.scratchpad.push_str(utf8ch.as_str());
+				Ok(ST(
+					State::Content(ContentState::Initial),
+					None,
+				))
+			},
+		}
+	}
+
 	fn lex_maybe_cdata_end<'r, R: CodepointRead>(&mut self, in_cdata: bool, nend: usize, r: &'r mut R) -> Result<ST> {
 		debug_assert!(nend < TOK_XML_CDATA_END.len());
 		let ctx = if in_cdata {
@@ -785,6 +831,7 @@ impl Lexer {
 		} else if ch == ']' {
 			// sequence was broken, but careful! this could just be `]]]]]]]>` sequence!
 			// those we need to treat the same, no matter whether inside or outside CDATA the previously found ] is moved to the scratchpad and we return to this state
+			self.prep_scratchpad();
 			self.scratchpad.push_str("]");
 			Ok(ST(
 				State::Content(ContentState::MaybeCDataEnd(in_cdata, nend)),
@@ -794,6 +841,7 @@ impl Lexer {
 			// sequence was broken
 			// the token string constant is utf8 (ascii even), we’re safe.
 			let encountered = unsafe { std::str::from_utf8_unchecked(&TOK_XML_CDATA_END[..nend]) };
+			self.prep_scratchpad();
 			self.scratchpad.push_str(encountered);
 			if in_cdata {
 				if CLASS_XML_NONCHAR.select(ch) {
@@ -801,6 +849,7 @@ impl Lexer {
 					Err(Error::NotWellFormed(WFError::InvalidChar(ERRCTX_CDATA_SECTION, ch as u32, false)))
 				} else {
 					// broken sequence inside cdata section, that’s fine; just push whatever we read to the scratchpad and move on
+					// no need for prep, we pushed above already
 					self.scratchpad.push_str(utf8ch.as_str());
 					Ok(ST(
 						State::Content(ContentState::CDataSection),
@@ -810,22 +859,7 @@ impl Lexer {
 				}
 			} else {
 				// broken sequence outside cdata section, need to analyze the next char carefully to handle entities and such
-				match self.lex_posttext_char(utf8ch)? {
-					// special delimiter char -> state transition
-					Some(st) => Ok(st),
-					// no special char -> check if it is valid text and proceed accordingly
-					None => if CLASS_XML_NONCHAR.select(ch) {
-						// non-Char, error
-						Err(Error::NotWellFormed(WFError::InvalidChar(ERRCTX_TEXT, ch as u32, false)))
-					} else {
-						// nothing special, push to scratchpad and return to initial content state
-						self.scratchpad.push_str(utf8ch.as_str());
-						Ok(ST(
-							State::Content(ContentState::Initial),
-							None,
-						))
-					},
-				}
+				self.lex_resume_text(utf8ch)
 			}
 		}
 	}
@@ -836,6 +870,62 @@ impl Lexer {
 			ContentState::MaybeElement(substate) => self.lex_maybe_element(substate, r),
 			ContentState::MaybeCDataEnd(in_cdata, nend) => self.lex_maybe_cdata_end(in_cdata, nend, r),
 
+			ContentState::MaybeCRLF(in_cdata) => {
+				let utf8ch = handle_eof(self.read_single(r)?, ERRCTX_TEXT)?;
+				match utf8ch.to_char() {
+					'\n' => {
+						// CRLF sequence, only insert the \n to the scratchpad.
+						self.prep_scratchpad();
+						self.scratchpad.push_str("\n");
+						// return to the content state and curse a bit
+						Ok(ST(
+							if in_cdata {
+								State::Content(ContentState::CDataSection)
+							} else {
+								State::Content(ContentState::Initial)
+							},
+							None,
+						))
+					},
+					'\r' => {
+						// double CR, so this may still be followed by an LF; but the first CR gets converted to LF
+						self.prep_scratchpad();
+						self.scratchpad.push_str("\n");
+						// stay in the same state, we may still get an LF here.
+						Ok(ST(
+							State::Content(ContentState::MaybeCRLF(in_cdata)),
+							None,
+						))
+					},
+					ch => {
+						// we read a single CR, so we push a \n to the scratchpad and hope for the best
+						self.prep_scratchpad();
+						self.scratchpad.push_str("\n");
+						if in_cdata {
+							// only special thing in CDATA is ']'
+							if ch == ']' {
+								Ok(ST(
+									State::Content(ContentState::MaybeCDataEnd(true, 1)),
+									None,
+								))
+							} else if !CLASS_XML_NONCHAR.select(ch) {
+								// ^ but of course we still need to check for a valid char. Thanks afl.
+								// no need for prep as we pushed above already
+								self.scratchpad.push_str(utf8ch.as_str());
+								Ok(ST(
+									State::Content(ContentState::CDataSection),
+									None,
+								))
+							} else{
+								Err(Error::NotWellFormed(WFError::InvalidChar(ERRCTX_CDATA_SECTION, ch as u32, false)))
+							}
+						} else {
+							self.lex_resume_text(utf8ch)
+						}
+					},
+				}
+			},
+
 			// read until next `<` or `&`, which are the only things which
 			// can break us out of this state.
 			ContentState::Initial => match self.read_validated(r, &CodepointRanges(VALID_XML_CDATA_RANGES_TEXT_DELIMITED), self.opts.max_token_length)? {
@@ -864,20 +954,26 @@ impl Lexer {
 					self.maybe_flush_scratchpad_as_text(0),
 				)),
 				// -> transition into the "first delimiter found" state
-				Endpoint::Delimiter(utf8ch) if utf8ch.to_char() == ']' => Ok(ST(
-					State::Content(ContentState::MaybeCDataEnd(true, 1)),
-					None,
-				)),
-				Endpoint::Delimiter(other) => Err(Error::NotWellFormed(WFError::InvalidChar(ERRCTX_CDATA_SECTION, other.to_char() as u32, false))),
+				Endpoint::Delimiter(utf8ch) => match utf8ch.to_char() {
+					']' => Ok(ST(
+						State::Content(ContentState::MaybeCDataEnd(true, 1)),
+						None,
+					)),
+					'\r' => Ok(ST(
+						State::Content(ContentState::MaybeCRLF(true)),
+						None,
+					)),
+					other => Err(Error::NotWellFormed(WFError::InvalidChar(ERRCTX_CDATA_SECTION, other as u32, false)))
+				}
 			},
-			ContentState::Whitespace => match self.skip_matching(r, &CLASS_XML_SPACES)? {
-				(_, Endpoint::Eof) | (_, Endpoint::Limit) => {
+			ContentState::Whitespace => match self.skip_matching(r, &CLASS_XML_SPACES) {
+				(_, Ok(Endpoint::Eof)) | (_, Ok(Endpoint::Limit)) => {
 					Ok(ST(
 						State::Eof,
 						None,
 					))
 				},
-				(_, Endpoint::Delimiter(ch)) => match ch.to_char() {
+				(_, Ok(Endpoint::Delimiter(ch))) => match ch.to_char() {
 					'<' => Ok(ST(
 						State::Content(ContentState::MaybeElement(MaybeElementState::Initial)), None,
 					)),
@@ -886,7 +982,8 @@ impl Lexer {
 						other,
 						Some(&["Spaces", "<"]),
 					))),
-				}
+				},
+				(_, Err(e)) => Err(e),
 			},
 		}
 	}
@@ -894,8 +991,8 @@ impl Lexer {
 	fn lex_element_postblank(&mut self, kind: ElementKind, utf8ch: Utf8Char) -> Result<ElementState> {
 		match utf8ch.to_char() {
 			' ' | '\t' | '\r' | '\n' => Ok(ElementState::Blank),
-			'"' => Ok(ElementState::AttributeValue('"', CodepointRanges(VALID_XML_CDATA_RANGES_ATT_QUOT_DELIMITED))),
-			'\'' => Ok(ElementState::AttributeValue('\'', CodepointRanges(VALID_XML_CDATA_RANGES_ATT_APOS_DELIMITED))),
+			'"' => Ok(ElementState::AttributeValue('"', CodepointRanges(VALID_XML_CDATA_RANGES_ATT_QUOT_DELIMITED), false)),
+			'\'' => Ok(ElementState::AttributeValue('\'', CodepointRanges(VALID_XML_CDATA_RANGES_ATT_APOS_DELIMITED), false)),
 			'=' => Ok(ElementState::Eq),
 			'>' => match kind {
 				ElementKind::Footer | ElementKind::Header => Ok(ElementState::Close),
@@ -912,6 +1009,7 @@ impl Lexer {
 			},
 			ch if CLASS_XML_NAMESTART.select(ch) => {
 				// write the char to scratchpad because it’ll be needed.
+				self.prep_scratchpad();
 				self.scratchpad.push_str(utf8ch.as_str());
 				Ok(ElementState::Name)
 			},
@@ -926,35 +1024,123 @@ impl Lexer {
 		}
 	}
 
+	fn lex_attval_next(&mut self, delim: char, selector: CodepointRanges, utf8ch: Utf8Char, element_kind: ElementKind) -> Result<ST> {
+		match utf8ch.to_char() {
+			'<' => Err(Error::NotWellFormed(WFError::UnexpectedChar(ERRCTX_ATTVAL, '<', None))),
+			'&' => {
+				// must swap scratchpad here to avoid clobbering the
+				// attribute value during entity read
+				self.swap_scratchpad()?;
+				Ok(ST(
+					State::Reference{
+						ctx: ERRCTX_ATTVAL,
+						ret: RefReturnState::AttributeValue(
+							element_kind,
+							delim,
+							selector,
+						),
+						kind: RefKind::Entity,
+					}, None
+				))
+			},
+			'\t' | '\n' => {
+				self.prep_scratchpad();
+				self.scratchpad.push_str(" ");
+				Ok(ST(
+					State::Element{
+						kind: element_kind,
+						state: ElementState::AttributeValue(delim, selector, false),
+					},
+					None,
+				))
+			},
+			'\r' => {
+				Ok(ST(
+					State::Element{
+						kind: element_kind,
+						state: ElementState::AttributeValue(delim, selector, true),
+					},
+					None,
+				))
+			},
+			d if d == delim => Ok(ST(
+				State::Element{
+					kind: element_kind,
+					// require whitespace after attribute as the grammar demands
+					state: ElementState::SpaceRequired,
+				},
+				Some(Token::AttributeValue(self.metrics(0), self.flush_scratchpad_as_cdata())),
+			)),
+			other => Err(Error::NotWellFormed(WFError::InvalidChar(
+				ERRCTX_ATTVAL,
+				other as u32,
+				false,
+			)))
+		}
+	}
+
 	fn lex_element<'r, R: CodepointRead>(&mut self, kind: ElementKind, state: ElementState, r: &'r mut R) -> Result<ST> {
 		match state {
-			ElementState::Start | ElementState::Name => match self.read_validated(r, &CLASS_XML_NAME, self.opts.max_token_length)? {
-				Endpoint::Eof => Err(Error::wfeof(ERRCTX_NAME)),
-				Endpoint::Limit => Err(self.token_length_error()),
-				Endpoint::Delimiter(ch) => {
-					let next_state = self.lex_element_postblank(kind, ch)?;
-					let name = self.flush_scratchpad_as_name();
-					let metrics = self.metrics(ch.len());
+			ElementState::Start | ElementState::Name => {
+				if self.scratchpad.len() == 0 {
+					// we are reading the first char; the first one is special because it must match CLASS_XML_NAMESTART, and not just CLASS_XML_NAME
+					let utf8ch = handle_eof(self.read_single(r)?, ERRCTX_NAME)?;
+					let ch = utf8ch.to_char();
+					if !CLASS_XML_NAMESTART.select(ch) {
+						Err(Error::NotWellFormed(WFError::UnexpectedChar(ERRCTX_NAME, ch, None)))
+					} else {
+						self.prep_scratchpad();
+						self.scratchpad.push_str(utf8ch.as_str());
+						// continue in the same state; the branch below will be taken next and read_validated will take care of it if we’re done already
+						Ok(ST(
+							State::Element{
+								kind: kind,
+								state: state,
+							},
+							None,
+						))
+					}
+				} else {
+					match self.read_validated(r, &CLASS_XML_NAME, self.opts.max_token_length)? {
+						Endpoint::Eof => Err(Error::wfeof(ERRCTX_NAME)),
+						Endpoint::Limit => Err(self.token_length_error()),
+						Endpoint::Delimiter(ch) => {
+							let next_state = self.lex_element_postblank(kind, ch)?;
+							let name = self.flush_scratchpad_as_name();
+							let metrics = self.metrics(ch.len());
+							Ok(ST(
+								State::Element{
+									kind: kind,
+									state: next_state,
+								},
+								Some(if state == ElementState::Name {
+									Token::Name(metrics, name)
+								} else {
+									match kind {
+										ElementKind::Header => Token::ElementHeadStart(metrics, name),
+										ElementKind::Footer => Token::ElementFootStart(metrics, name),
+										ElementKind::XMLDecl => panic!("invalid state"),
+									}
+								}),
+							))
+						}
+					}
+				}
+			},
+			ElementState::SpaceRequired | ElementState::Blank => match self.skip_matching(r, &CLASS_XML_SPACES) {
+				(_, Ok(Endpoint::Eof)) | (_, Ok(Endpoint::Limit)) => Err(Error::wfeof(ERRCTX_ELEMENT)),
+				(nmatching, Err(Error::IO(_))) if nmatching > 0 && state == ElementState::SpaceRequired => {
+					// we have to treat IO errors here specially and implicitly retry them (because we swallow this one). that is in line with the contract which says that IO errors are retriable
+					// this is because we need to transition from SpaceRequired to Blank after reading even only a single char. otherwise, we are not resilient against chunking.
 					Ok(ST(
 						State::Element{
 							kind: kind,
-							state: next_state,
+							state: ElementState::Blank,
 						},
-						Some(if state == ElementState::Name {
-							Token::Name(metrics, name)
-						} else {
-							match kind {
-								ElementKind::Header => Token::ElementHeadStart(metrics, name),
-								ElementKind::Footer => Token::ElementFootStart(metrics, name),
-								ElementKind::XMLDecl => panic!("invalid state"),
-							}
-						}),
+						None,
 					))
 				},
-			},
-			ElementState::SpaceRequired | ElementState::Blank => match self.skip_matching(r, &CLASS_XML_SPACES)? {
-				(_, Endpoint::Eof) | (_, Endpoint::Limit) => Err(Error::wfeof(ERRCTX_ELEMENT)),
-				(nmatching, Endpoint::Delimiter(ch)) => {
+				(nmatching, Ok(Endpoint::Delimiter(ch))) => {
 					self.eat_whitespace_metrics(ch.len());
 					let next_state = self.lex_element_postblank(kind, ch)?;
 					if next_state == ElementState::Name && state == ElementState::SpaceRequired && nmatching == 0 {
@@ -971,43 +1157,32 @@ impl Lexer {
 						))
 					}
 				},
+				(_, Err(e)) => Err(e),
 			},
 			// XML 1.0 §2.3 [10] AttValue
-			ElementState::AttributeValue(delim, selector) => match self.read_validated(r, &selector, self.opts.max_token_length)? {
+			ElementState::AttributeValue(delim, selector, false) => match self.read_validated(r, &selector, self.opts.max_token_length)? {
 				Endpoint::Eof => Err(Error::wfeof(ERRCTX_ATTVAL)),
 				Endpoint::Limit => Err(self.token_length_error()),
-				Endpoint::Delimiter(utf8ch) => match utf8ch.to_char() {
-					'<' => Err(Error::NotWellFormed(WFError::UnexpectedChar(ERRCTX_ATTVAL, '<', None))),
-					'&' => {
-						// must swap scratchpad here to avoid clobbering the
-						// attribute value during entity read
-						self.swap_scratchpad()?;
-						Ok(ST(
-							State::Reference{
-								ctx: ERRCTX_ATTVAL,
-								ret: RefReturnState::AttributeValue(
-									kind,
-									delim,
-									selector,
-								),
-								kind: RefKind::Entity,
-							}, None
-						))
-					},
-					d if d == delim => Ok(ST(
+				Endpoint::Delimiter(utf8ch) => self.lex_attval_next(delim, selector, utf8ch, kind),
+			},
+			// CRLF normalization for attributes; cannot reuse the element mechanism here because we have to carry around the delimiter and stuff
+			ElementState::AttributeValue(delim, selector, true) => {
+				let utf8ch = handle_eof(self.read_single(r)?, ERRCTX_ATTVAL)?;
+				if utf8ch.to_char() == '\r' {
+					// push the space, continue with CRLF
+					self.prep_scratchpad();
+					self.scratchpad.push_str(" ");
+					Ok(ST(
 						State::Element{
 							kind: kind,
-							// require whitespace after attribute as the grammar demands
-							state: ElementState::SpaceRequired,
+							state: ElementState::AttributeValue(delim, selector, true),
 						},
-						Some(Token::AttributeValue(self.metrics(0), self.flush_scratchpad_as_cdata())),
-					)),
-					other => Err(Error::NotWellFormed(WFError::InvalidChar(
-						ERRCTX_ATTVAL,
-						other as u32,
-						false,
-					)))
-				},
+						None,
+					))
+				} else {
+					// not another CR, so we can move on to the default handling
+					self.lex_attval_next(delim, selector, utf8ch, kind)
+				}
 			},
 			ElementState::MaybeXMLDeclEnd => match self.read_single(r)? {
 				Some(ch) if ch.to_char() == '>' => {
@@ -1145,7 +1320,7 @@ impl Lexer {
 	pub fn lex<'r, R: CodepointRead>(&mut self, r: &'r mut R) -> Result<Option<Token>>
 	{
 		if let Some(e) = self.err.as_ref() {
-			return Err(e.clone())
+			return Err((**e).clone())
 		}
 
 		loop {
@@ -1163,7 +1338,7 @@ impl Lexer {
 				Err(other) => {
 					// we cache all other errors because we don't want to read / emit invalid data
 					let err = other.clone();
-					self.err = Some(other);
+					self.err = Some(Box::new(other));
 					return Err(err);
 				},
 				Ok(st) => st,
@@ -1191,6 +1366,16 @@ impl Lexer {
 			}
 		}
 	}
+
+	/// Release all temporary buffers
+	///
+	/// This is sensible to call when it is expected that no more data will be
+	/// processed by the lexer for a while and the memory is better used
+	/// elsewhere.
+	pub fn release_temporaries(&mut self) {
+		self.scratchpad.shrink_to_fit();
+		self.swap.shrink_to_fit();
+	}
 }
 
 impl fmt::Debug for Lexer {
@@ -1219,12 +1404,8 @@ mod tests {
 	fn stream_to_sink<'r, 's, 'l, R: CodepointRead, S: Sink>(l: &'l mut Lexer, r: &'r mut R, s: &'s mut S) -> Result<()> {
 		loop {
 			match l.lex(r) {
-				Ok(Some(tok)) => {
-					s.token(tok);
-				},
-				Ok(None) => {
-					break;
-				},
+				Ok(Some(tok)) => s.token(tok),
+				Ok(None) => break,
 				Err(e) => return Err(e),
 			}
 		}
@@ -1296,6 +1477,23 @@ mod tests {
 		(sink.dest, result)
 	}
 
+	fn lex_chunked(data: &[&[u8]], token_limit: usize) -> (Vec<Token>, Result<()>) {
+		let mut buff = BufferQueue::new();
+		let mut src = DecodingReader::new(&mut buff);
+		let mut lexer = Lexer::new();
+		let mut sink = VecSink::new(token_limit);
+		for chunk in data.iter() {
+			src.get_mut().push(*chunk);
+			match stream_to_sink(&mut lexer, &mut src, &mut sink) {
+				Ok(()) => panic!("unexpected end of tokens"),Err(Error::IO(ioerr)) if ioerr.kind() == io::ErrorKind::WouldBlock => (),
+				Err(e) => return (sink.dest, Err(e)),
+			}
+		}
+		src.get_mut().push_eof();
+		let result = stream_to_sink(&mut lexer, &mut src, &mut sink);
+		(sink.dest, result)
+	}
+
 	fn lex_err(data: &[u8], token_limit: usize) -> Option<Error> {
 		let (_, r) = lex(data, token_limit);
 		r.err()
@@ -2048,6 +2246,20 @@ mod tests {
 		assert!(matches!(err, Error::NotWellFormed(WFError::InvalidChar(_, 0u32, false))));
 	}
 
+	#[test]
+	fn lexer_rejects_numeric_start_of_name_in_closing_tag() {
+		// found via fuzzing by moparisthebest
+		let err = lex_err(b"</4foo>", 128).unwrap();
+		assert!(matches!(err, Error::NotWellFormed(WFError::UnexpectedChar(_, '4', None))));
+	}
+
+	#[test]
+	fn lexer_rejects_zero_length_name_in_closing_tag() {
+		// found via fuzzing by moparisthebest
+		let err = lex_err(b"</ >", 128).unwrap();
+		assert!(matches!(err, Error::NotWellFormed(WFError::UnexpectedChar(_, ' ', None))));
+	}
+
 	#[test]
 	fn lexer_lex_accounts_whitespace_between_xml_decl_and_element_to_element() {
 		// even though this behaviour may seem strange, it is useful for making sure that all bytes are accounted for in the parser.
@@ -2075,4 +2287,220 @@ mod tests {
 			other => panic!("unexpected event: {:?}", other),
 		}
 	}
+
+	#[test]
+	fn lexer_folds_crlf_to_lf_in_text() {
+		// XML 1.0 § 2.11
+		let (toks, r) = lex(b"<a>\r\n</a>", 128);
+		r.unwrap();
+
+		let mut iter = toks.iter();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		match iter.next().unwrap() {
+			Token::Text(_, cdata) => {
+				assert_eq!(cdata, "\n");
+			},
+			other => panic!("unexpected token: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn lexer_rejects_nonchar_after_cr() {
+		let err = lex_err(b"<a>\r\x01</a>", 128).unwrap();
+		match err {
+			Error::NotWellFormed(WFError::InvalidChar(_, 1, false)) => (),
+			other => panic!("unexpected error: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn lexer_rejects_nonchar_after_cr_in_cdata() {
+		// found with afl
+		let err = lex_err(b"<a><![CDATA[\r\x01]]></a>", 128).unwrap();
+		match err {
+			Error::NotWellFormed(WFError::InvalidChar(_, 1, false)) => (),
+			other => panic!("unexpected error: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn lexer_does_not_modify_charrefs_for_line_endings() {
+		// XML 1.0 § 2.11
+		let (toks, r) = lex(b"<a>&#xd;&#xa;</a>", 128);
+		r.unwrap();
+
+		let mut iter = toks.iter();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		let (text, _, _, _) = collect_texts(&mut iter);
+		assert_eq!(text, "\r\n");
+	}
+
+	#[test]
+	fn lexer_folds_crlf_to_lf_in_cdata() {
+		// XML 1.0 § 2.11
+		let (toks, r) = lex(b"<a><![CDATA[\r\n]]></a>", 128);
+		r.unwrap();
+
+		let mut iter = toks.iter();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		match iter.next().unwrap() {
+			Token::Text(_, cdata) => {
+				assert_eq!(cdata, "\n");
+			},
+			other => panic!("unexpected token: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn lexer_cr_folding_does_not_break_specials() {
+		// XML 1.0 § 2.11
+		let (toks, r) = lex(b"<a>\r</a>", 128);
+		r.unwrap();
+
+		let mut iter = toks.iter();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		match iter.next().unwrap() {
+			Token::Text(_, cdata) => {
+				assert_eq!(cdata, "\n");
+			},
+			other => panic!("unexpected token: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn lexer_cr_folding_in_cdata_does_not_break_exit() {
+		// XML 1.0 § 2.11
+		let (toks, r) = lex(b"<a><![CDATA[\r]]></a>", 128);
+		r.unwrap();
+
+		let mut iter = toks.iter();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		match iter.next().unwrap() {
+			Token::Text(_, cdata) => {
+				assert_eq!(cdata, "\n");
+			},
+			other => panic!("unexpected token: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn lexer_cr_folding_in_cdata_does_not_break_exit_cdata_section() {
+		// XML 1.0 § 2.11
+		let (toks, r) = lex(b"<a><![CDATA[\r<>]]></a>", 128);
+		r.unwrap();
+
+		let mut iter = toks.iter();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		match iter.next().unwrap() {
+			Token::Text(_, cdata) => {
+				assert_eq!(cdata, "\n<>");
+			},
+			other => panic!("unexpected token: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn lexer_folds_crcrlf_to_lflf_in_text() {
+		// XML 1.0 § 2.11
+		let (toks, r) = lex(b"<a>\r\r\n</a>", 128);
+		r.unwrap();
+
+		let mut iter = toks.iter();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		match iter.next().unwrap() {
+			Token::Text(_, cdata) => {
+				assert_eq!(cdata, "\n\n");
+			},
+			other => panic!("unexpected token: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn lexer_folds_cr_to_lf_in_text() {
+		// XML 1.0 § 2.11
+		let (toks, r) = lex(b"<a>\r</a>", 128);
+		r.unwrap();
+
+		let mut iter = toks.iter();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		match iter.next().unwrap() {
+			Token::Text(_, cdata) => {
+				assert_eq!(cdata, "\n");
+			},
+			other => panic!("unexpected token: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn lexer_normalizes_whitespace_in_attributes() {
+		// XML 1.0 § 3.3.3
+		let (toks, r) = lex(b"<a x='\r\r\n\t '/>", 128);
+		r.unwrap();
+
+		let mut iter = toks.iter();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		match iter.next().unwrap() {
+			Token::AttributeValue(_, cdata) => {
+				// just four spaces, because CRLF normalization happens before attribute value normalization
+				// gotta love this
+				assert_eq!(cdata, "    ");
+			},
+			other => panic!("unexpected token: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn lexer_handles_crlf_in_attribute() {
+		// XML 1.0 § 3.3.3
+		let (toks, r) = lex(b"<a x='\r\n\t '/>", 128);
+		r.unwrap();
+
+		let mut iter = toks.iter();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		match iter.next().unwrap() {
+			Token::AttributeValue(_, cdata) => {
+				assert_eq!(cdata, "   ");
+			},
+			other => panic!("unexpected token: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn lexer_preserves_whitespace_inserted_via_charrefs_in_attributes() {
+		// XML 1.0 § 3.3.3
+		let (toks, r) = lex(b"<a x='&#xd;&#xa;&#x9; '/>", 128);
+		r.unwrap();
+
+		let mut iter = toks.iter();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		iter.next().unwrap();
+		match iter.next().unwrap() {
+			Token::AttributeValue(_, cdata) => {
+				assert_eq!(cdata, "\r\n\t ");
+			},
+			other => panic!("unexpected token: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn lexer_is_resilient_to_chunking() {
+		let (_toks, r) = lex_chunked(
+			&[&b"<foo bar='baz' "[..], &b"fnord=''/>"[..]],
+			128,
+		);
+		r.unwrap();
+	}
 }
