@@ -54,6 +54,7 @@ pub mod parser;
 mod bufq;
 pub mod strings;
 mod selectors;
+mod context;
 
 #[cfg(test)]
 mod tests;
@@ -63,10 +64,13 @@ pub use error::{Error, Result};
 #[doc(inline)]
 pub use lexer::{Lexer, LexerOptions, DecodingReader};
 #[doc(inline)]
-pub use parser::{Parser, Event, LexerAdapter, XMLVersion};
+pub use parser::{QName, Parser, Event, LexerAdapter, XMLVersion, XMLNS_XML};
 #[doc(inline)]
 pub use bufq::BufferQueue;
 pub use strings::{NCName, Name, NCNameStr, NameStr, CData, CDataStr};
+pub use context::Context;
+
+pub const VERSION: &'static str = env!("CARGO_PKG_VERSION");
 
 /**
 # Source for individual XML events
@@ -178,12 +182,16 @@ pub fn as_eof_flag(r: Result<()>) -> Result<bool> {
 impl<'x> FeedParser<'x> {
 	/// Create a new default `FeedParser`.
 	pub fn new() -> FeedParser<'x> {
+		Self::with_context(parser::RcPtr::new(Context::new()))
+	}
+
+	pub fn with_context(ctx: parser::RcPtr<Context>) -> FeedParser<'x> {
 		FeedParser{
 			token_source: LexerAdapter::new(
 				Lexer::new(),
 				DecodingReader::new(BufferQueue::new()),
 			),
-			parser: Parser::new(),
+			parser: Parser::with_context(ctx),
 		}
 	}
 
@@ -225,6 +233,24 @@ impl<'x> FeedParser<'x> {
 	pub fn buffered(&self) -> usize {
 		self.token_source.get_ref().get_ref().len()
 	}
+
+	/// Return a reference to the internal buffer BufferQueue
+	///
+	/// This can be used to force dropping of all memory in case of error
+	/// conditions.
+	pub fn get_buffer_mut(&mut self) -> &mut BufferQueue<'x> {
+		self.token_source.get_mut().get_mut()
+	}
+
+	/// Release all temporary buffers
+	///
+	/// This is sensible to call when it is expected that no more data will be
+	/// processed by the parser for a while and the memory is better used
+	/// elsewhere.
+	pub fn release_temporaries(&mut self) {
+		self.token_source.get_lexer_mut().release_temporaries();
+		self.parser.release_temporaries();
+	}
 }
 
 impl EventRead for FeedParser<'_> {
