@@ -13,10 +13,11 @@ use std::collections::VecDeque;
 use crate::lexer::{Token, Lexer, CodepointRead, TokenMetrics};
 use crate::error::*;
 use crate::strings::*;
+use crate::context;
 
-pub const XML_NAMESPACE: &'static CDataStr = unsafe { std::mem::transmute("http://www.w3.org/XML/1998/namespace") };
+pub const XMLNS_XML: &'static CDataStr = unsafe { std::mem::transmute("http://www.w3.org/XML/1998/namespace") };
 
-type QName = (Option<RcPtr<CData>>, NCName);
+pub type QName = (Option<RcPtr<CData>>, NCName);
 
 /// Wrapper pointer around namespace URIs
 ///
@@ -69,13 +70,14 @@ impl EventMetrics {
 		self.len
 	}
 
-	// for use in crate unit tests
-	#[allow(dead_code)]
-	pub(crate) const fn new(len: usize) -> EventMetrics {
+	// Create new event metrics
+	pub const fn new(len: usize) -> EventMetrics {
 		EventMetrics{len: len}
 	}
 }
 
+pub static ZERO_METRICS: EventMetrics = EventMetrics::new(0);
+
 /**
 # XML document parts
 
@@ -119,6 +121,18 @@ pub enum Event {
 	Text(EventMetrics, CData),
 }
 
+impl Event {
+	/// Return the [`EventMetrics`] of the event
+	pub fn metrics(&self) -> &EventMetrics {
+		match self {
+			Self::XMLDeclaration(m, ..) => &m,
+			Self::StartElement(m, ..) => &m,
+			Self::EndElement(m, ..) => &m,
+			Self::Text(m, ..) => &m,
+		}
+	}
+}
+
 #[derive(Clone, Copy, PartialEq, Debug)]
 enum DeclSt {
 	VersionName,
@@ -202,6 +216,7 @@ It is a low-level interface which expects to be driven from a [`TokenRead`]
 source.
 */
 pub struct Parser {
+	ctx: RcPtr<context::Context>,
 	state: State,
 	fixed_xml_namespace: RcPtr<CData>,
 	/// keep a stack of the element Names (i.e. (Prefix:)?Localname) as a
@@ -220,15 +235,21 @@ pub struct Parser {
 	/// multiple events need to be pushed from a single token, which is why
 	/// the queue exists as a buffer.
 	eventq: VecDeque<Event>,
-	err: Option<Error>,
+	err: Option<Box<Error>>,
 }
 
 impl Parser {
 	/// Create a new parser
 	pub fn new() -> Parser {
+		Self::with_context(RcPtr::new(context::Context::new()))
+	}
+
+	pub fn with_context(ctx: RcPtr<context::Context>) -> Parser {
+		let xmlns = ctx.intern_cdata(XMLNS_XML.to_cdata());
 		Parser{
+			ctx: ctx,
 			state: State::Initial,
-			fixed_xml_namespace: RcPtr::new(XML_NAMESPACE.to_cdata()),
+			fixed_xml_namespace: xmlns,
 			element_stack: Vec::new(),
 			namespace_stack: Vec::new(),
 			element_scratchpad: None,
@@ -287,13 +308,13 @@ impl Parser {
 
 	/// Poison the parser, making it return the same error for all eternity.
 	fn poison(&mut self, e: Error) -> () {
-		self.err = Some(e)
+		self.err = Some(Box::new(e))
 	}
 
 	/// Check if the parser is poisoned and return the corresponding error.
 	fn check_poison(&self) -> Result<()> {
 		if let Some(e) = self.err.as_ref() {
-			Err(e.clone())
+			Err((**e).clone())
 		} else {
 			Ok(())
 		}
@@ -555,14 +576,14 @@ impl Parser {
 				if localname == "xmlns" {
 					Err(Error::NotNamespaceWellFormed(NWFError::ReservedNamespacePrefix))
 				} else if localname == "xml" {
-					if val != XML_NAMESPACE {
+					if val != XMLNS_XML {
 						Err(Error::NotNamespaceWellFormed(NWFError::ReservedNamespacePrefix))
 					} else {
 						Ok(())
 					}
 				} else if val.len() == 0 {
 					Err(Error::NotNamespaceWellFormed(NWFError::EmptyNamespaceUri))
-				} else if scratchpad.namespace_decls.insert(localname, RcPtr::new(val)).is_some() {
+				} else if scratchpad.namespace_decls.insert(localname, self.ctx.intern_cdata(val)).is_some() {
 					Err(Error::NotWellFormed(WFError::DuplicateAttribute))
 				} else {
 					Ok(())
@@ -573,7 +594,7 @@ impl Parser {
 				if scratchpad.default_namespace_decl.is_some() {
 					Err(Error::NotWellFormed(WFError::DuplicateAttribute))
 				} else {
-					scratchpad.default_namespace_decl = Some(RcPtr::new(val));
+					scratchpad.default_namespace_decl = Some(self.ctx.intern_cdata(val));
 					Ok(())
 				}
 			},
@@ -735,6 +756,8 @@ impl Parser {
 				State::Document(substate) => self.parse_document(substate, r),
 				State::End => match self.read_token(r)? {
 					None => Ok(State::Eof),
+					// whitespace after the root element is explicitly allowed
+					Some(Token::Text(_, s)) if s.as_bytes().iter().all(|&c| c == b' ' || c == b'\t' || c == b'\n' || c == b'\r') => Ok(State::End),
 					Some(tok) => Err(Error::NotWellFormed(WFError::UnexpectedToken(
 						ERRCTX_DOCEND,
 						tok.name(),
@@ -755,6 +778,17 @@ impl Parser {
 			};
 		}
 	}
+
+	/// Release all temporary buffers
+	///
+	/// This is sensible to call when it is expected that no more data will be
+	/// processed by the parser for a while and the memory is better used
+	/// elsewhere.
+	pub fn release_temporaries(&mut self) {
+		self.eventq.shrink_to_fit();
+		self.element_stack.shrink_to_fit();
+		self.namespace_stack.shrink_to_fit();
+	}
 }
 
 impl fmt::Debug for Parser {
@@ -795,6 +829,16 @@ impl<R: CodepointRead + Sized> LexerAdapter<R> {
 	pub fn get_mut(&mut self) -> &mut R {
 		&mut self.src
 	}
+
+	/// Return a reference to the lexer
+	pub fn get_lexer(&mut self) -> &Lexer {
+		&self.lexer
+	}
+
+	/// Return a mutable reference to the lexer
+	pub fn get_lexer_mut(&mut self) -> &mut Lexer {
+		&mut self.lexer
+	}
 }
 
 impl<R: CodepointRead + Sized> TokenRead for LexerAdapter<R> {
@@ -808,6 +852,7 @@ mod tests {
 	use super::*;
 	use std::io;
 	use crate::lexer::TokenMetrics;
+	use std::convert::TryInto;
 
 	const TEST_NS: &'static str = "urn:uuid:4e1c8b65-ae37-49f8-a250-c27d52827da9";
 	const TEST_NS2: &'static str = "urn:uuid:678ba034-6200-4ecd-803f-bbcbfa225236";
@@ -1478,4 +1523,62 @@ mod tests {
 		let err = parse_err(toks).unwrap();
 		assert!(matches!(err, Error::NotNamespaceWellFormed(NWFError::UndeclaredNamesacePrefix(_))));
 	}
+
+	#[test]
+	fn parser_reject_element_after_root_element() {
+		let (evs, r) = parse(&[
+			Token::ElementHeadStart(DM, Name::from_str("root").unwrap()),
+			Token::ElementHFEnd(DM),
+			Token::ElementFootStart(DM, Name::from_str("root").unwrap()),
+			Token::ElementHFEnd(DM),
+			Token::ElementHeadStart(DM, Name::from_str("garbage").unwrap()),
+			Token::ElementHFEnd(DM),
+			Token::ElementFootStart(DM, Name::from_str("garbage").unwrap()),
+			Token::ElementHFEnd(DM),
+		]);
+		let mut iter = evs.iter();
+		assert!(matches!(iter.next().unwrap(), Event::StartElement(EventMetrics{len: 0}, (nsuri, localname), _attrs) if nsuri.is_none() && localname == "root"));
+		assert!(matches!(iter.next().unwrap(), Event::EndElement(EventMetrics{len: 0})));
+		assert!(iter.next().is_none());
+		match r {
+			Err(Error::NotWellFormed(WFError::UnexpectedToken(_, _, _))) => (),
+			other => panic!("unexpected result: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn parser_reject_text_after_root_element() {
+		let (evs, r) = parse(&[
+			Token::ElementHeadStart(DM, Name::from_str("root").unwrap()),
+			Token::ElementHFEnd(DM),
+			Token::ElementFootStart(DM, Name::from_str("root").unwrap()),
+			Token::ElementHFEnd(DM),
+			Token::Text(DM, "foo".try_into().unwrap()),
+		]);
+		let mut iter = evs.iter();
+		assert!(matches!(iter.next().unwrap(), Event::StartElement(EventMetrics{len: 0}, (nsuri, localname), _attrs) if nsuri.is_none() && localname == "root"));
+		assert!(matches!(iter.next().unwrap(), Event::EndElement(EventMetrics{len: 0})));
+		assert!(iter.next().is_none());
+		match r {
+			Err(Error::NotWellFormed(WFError::UnexpectedToken(_, _, _))) => (),
+			other => panic!("unexpected result: {:?}", other),
+		}
+	}
+
+	#[test]
+	fn parser_allow_whitespace_after_root_element() {
+		let (evs, r) = parse(&[
+			Token::ElementHeadStart(DM, Name::from_str("root").unwrap()),
+			Token::ElementHFEnd(DM),
+			Token::ElementFootStart(DM, Name::from_str("root").unwrap()),
+			Token::ElementHFEnd(DM),
+			Token::Text(DM, " \t\r\n".try_into().unwrap()),
+			Token::Text(DM, "\n\r\t ".try_into().unwrap()),
+		]);
+		let mut iter = evs.iter();
+		assert!(matches!(iter.next().unwrap(), Event::StartElement(EventMetrics{len: 0}, (nsuri, localname), _attrs) if nsuri.is_none() && localname == "root"));
+		assert!(matches!(iter.next().unwrap(), Event::EndElement(EventMetrics{len: 0})));
+		assert!(iter.next().is_none());
+		r.unwrap();
+	}
 }
