@@ -49,8 +49,9 @@ pub const VALID_XML_CDATA_RANGES: &'static [CodepointRange] = &[
 
 // XML 1.0 § 2.2
 pub const INVALID_XML_CDATA_RANGES: &'static [CodepointRange] = &[
-	CodepointRange('\x00', '\x09'),
+	CodepointRange('\x00', '\x08'),
 	CodepointRange('\x0b', '\x0c'),
+	CodepointRange('\x0e', '\x1f'),
 	CodepointRange('\u{fffe}', '\u{ffff}'),
 ];
 
@@ -58,7 +59,7 @@ pub const INVALID_XML_CDATA_RANGES: &'static [CodepointRange] = &[
 // XML 1.0 § 2.4 [14]
 pub const VALID_XML_CDATA_RANGES_TEXT_DELIMITED: &'static [CodepointRange] = &[
 	CodepointRange('\x09', '\x0a'),
-	CodepointRange('\x0d', '\x0d'),
+	// excluding CR as that gets folded to LF
 	CodepointRange('\u{0020}', '\u{0025}'), // excludes &
 	CodepointRange('\u{0027}', '\u{003b}'), // excludes <
 	CodepointRange('\u{003d}', '\u{005c}'), // excludes ]
@@ -70,8 +71,7 @@ pub const VALID_XML_CDATA_RANGES_TEXT_DELIMITED: &'static [CodepointRange] = &[
 
 // XML 1.0 § 2.3 [10]
 pub const VALID_XML_CDATA_RANGES_ATT_APOS_DELIMITED: &'static [CodepointRange] = &[
-	CodepointRange('\x09', '\x0a'),
-	CodepointRange('\x0d', '\x0d'),
+	// exclude all whitespace except normal space because those get converted into spaces
 	CodepointRange('\u{0020}', '\u{0025}'), // excludes &, '
 	CodepointRange('\u{0028}', '\u{003b}'), // excludes <
 	CodepointRange('\u{003d}', '\u{d7ff}'),
@@ -82,8 +82,7 @@ pub const VALID_XML_CDATA_RANGES_ATT_APOS_DELIMITED: &'static [CodepointRange] =
 
 // XML 1.0 § 2.3 [10]
 pub const VALID_XML_CDATA_RANGES_ATT_QUOT_DELIMITED: &'static [CodepointRange] = &[
-	CodepointRange('\x09', '\x0a'),
-	CodepointRange('\x0d', '\x0d'),
+	// exclude all whitespace except normal space because those get converted into spaces
 	CodepointRange('\u{0020}', '\u{0021}'), // excludes "
 	CodepointRange('\u{0023}', '\u{0025}'), // excludes &
 	CodepointRange('\u{0027}', '\u{003b}'), // excludes <
@@ -96,7 +95,7 @@ pub const VALID_XML_CDATA_RANGES_ATT_QUOT_DELIMITED: &'static [CodepointRange] =
 // XML 1.0 § 2.4 [14]
 pub const VALID_XML_CDATA_RANGES_CDATASECTION_DELIMITED: &'static [CodepointRange] = &[
 	CodepointRange('\x09', '\x0a'),
-	CodepointRange('\x0d', '\x0d'),
+	// excluding CR as that gets folded to LF
 	CodepointRange('\u{0020}', '\u{005c}'), // excludes ]
 	CodepointRange('\u{005e}', '\u{d7ff}'),
 	CodepointRange('\u{e000}', '\u{fffd}'),
@@ -211,3 +210,21 @@ impl PartialEq for CodepointRanges {
 		std::ptr::eq(&self.0, &other.0)
 	}
 }
+
+#[cfg(test)]
+mod tests {
+	use super::*;
+
+	#[test]
+	fn cdata_inclusion_and_exclusion_are_equivalent() {
+		let excluder = CodepointRanges(INVALID_XML_CDATA_RANGES);
+		let includer = CodepointRanges(VALID_XML_CDATA_RANGES);
+		for cp in 0x0..=0x10ffffu32 {
+			if let Some(ch) = std::char::from_u32(cp) {
+				if !includer.select(ch) != excluder.select(ch) {
+					panic!("INVALID_XML_CDATA_RANGES and VALID_XML_CDATA_RANGES have different opinions about U+{:x}", cp)
+				}
+			}
+		}
+	}
+}
